import rq
from redis import Redis

from judoscale.core.adapter import Adapter, AdapterInfo
from judoscale.core.config import config as judoconfig
from judoscale.core.reporter import reporter
from judoscale.rq.collector import RQMetricsCollector


def judoscale_rq(redis: Redis, extra_config: dict = {}) -> None:
    judoconfig.merge(extra_config)
    collector = RQMetricsCollector(config=judoconfig, redis=redis)
    adapter = Adapter(
        identifier="judoscale-rq",
        adapter_info=AdapterInfo(platform_version=rq.VERSION),
        metrics_collector=collector,
    )

    reporter.add_adapter(adapter)
    reporter.ensure_running()



import json
import time
from threading import Thread
from typing import List, Optional, Set

from celery import Celery
from kombu import Connection
from redis import Redis

from judoscale.core.config import Config
from judoscale.core.logger import logger
from judoscale.core.metric import Metric
from judoscale.core.metrics_collectors import JobMetricsCollector


class TaskSentHandler(Thread):
    def __init__(
        self,
        collector: "CeleryMetricsCollector",
        connection: Connection,
        *args,
        **kwargs,
    ):
        self.collector = collector
        self.connection = connection
        super().__init__(*args, daemon=True, **kwargs)

    def task_sent(self, event):
        self.collector.queues.add(event["queue"])

    def run(self):
        logger.debug("Starting TaskSentHandler")
        recv = self.collector.broker.events.Receiver(
            self.connection,
            handlers={"task-sent": self.task_sent},
        )
        recv.capture(wakeup=False)


class CeleryMetricsCollector(JobMetricsCollector):
    def __init__(self, config: Config, broker: Celery):
        super().__init__(config=config)

        self.broker = broker
        connection = self.broker.connection_for_read()
        if connection.transport.driver_name != "redis":
            raise NotImplementedError(
                f"Unsupported broker: {connection.transport.driver_name}"
            )

        self.redis: Redis = connection.channel().client
        self.queues: Set[str] = set()
        self.task_sent_handler = TaskSentHandler(self, connection)
        logger.debug(f"Redis is at {self.redis.connection_pool}")

        system_queues = {"unacked", "unacked_index"}
        user_queues = {
            q.decode("utf-8") if type(q) is bytes else q
            for q in self.redis.scan_iter(match="[^_]*", _type="list")
        }
        logger.debug(f"Found initial queues: {list(user_queues)}")
        self.queues = user_queues - system_queues
        self.task_sent_handler.start()

    def oldest_task(self, queue: str) -> Optional[dict]:
        """
        Get the oldest task from the queue.
        """
        try:
            if payload := self.redis.lindex(queue, -1):
                return json.loads(payload)
        except Exception as e:
            logger.warning(f"Unable to get a task from queue: {queue}", exc_info=e)
        return None

    def collect(self) -> List[Metric]:
        metrics = []
        if not self.should_collect:
            return metrics

        logger.debug(f"Collecting metrics for queues {list(self.queues)}")
        for queue in self.queues:
            if task := self.oldest_task(queue):
                if published_at := task["properties"].get("published_at"):
                    metrics.append(
                        Metric.for_queue(queue_name=queue, oldest_job_ts=published_at)
                    )
            else:
                metrics.append(
                    Metric.for_queue(queue_name=queue, oldest_job_ts=time.time())
                )

        return metrics



import logging
import os

from judoscale.core.logger import logger


class Config:
    def __init__(self):
        self.log_level = os.environ.get("LOG_LEVEL", "INFO")
        self.dyno = os.environ.get("DYNO", "none.0")
        self.dyno_name, self.dyno_num = self.dyno.split(".")
        self.dyno_num = int(self.dyno_num)
        self.report_interval_seconds = 10
        self.api_base_url = os.environ.get("JUDOSCALE_URL")
        self._prepare_logging()

    def merge(self, settings: dict):
        for key, value in settings.items():
            setattr(self, key.lower(), value)
        self._prepare_logging()

    def for_report(self):
        # Only include the config options we want to include in the report
        return {
            "log_level": self.log_level,
            "report_interval_seconds": self.report_interval_seconds,
        }

    def _prepare_logging(self):
        log_level = logging.getLevelName(self.log_level.upper())
        logger.setLevel(log_level)

        if not logger.handlers:
            stdout_handler = logging.StreamHandler()
            fmt = "%(levelname)s - [%(name)s] %(message)s"
            stdout_handler.setFormatter(logging.Formatter(fmt))
            logger.addHandler(stdout_handler)

        logger.propagate = False


config = Config()



from typing import List, Protocol

from judoscale.core.config import Config
from judoscale.core.metric import Metric
from judoscale.core.metrics_store import MetricsStore


class Collector(Protocol):
    def collect(self) -> List[Metric]:
        ...

    @property
    def should_collect(self) -> bool:
        ...


class MetricsCollector:
    def __init__(self, config: Config):
        self.config = config

    @property
    def should_collect(self) -> bool:
        return True


class WebMetricsCollector(MetricsCollector):
    def __init__(self, config: Config):
        self.store = MetricsStore()
        super().__init__(config=config)

    @property
    def should_collect(self):
        return self.config.dyno_name == "web"

    def add(self, metric: Metric):
        """
        Add metric to the store if it should be collected.
        """
        if self.should_collect:
            self.store.add(metric)

    def collect(self) -> List[Metric]:
        """
        Return all metrics in the store and clear the store.
        """
        return self.store.flush()


class JobMetricsCollector(MetricsCollector):
    def __init__(self, config: Config):
        super().__init__(config=config)

    @property
    def should_collect(self):
        return self.config.dyno_num == 1



import re
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional, Tuple

from judoscale.core.logger import logger


@dataclass
class Metric:
    timestamp: float  # Unix timestamp in fractional seconds
    value: int
    queue_name: Optional[str] = None
    measurement: str = "queue_time"

    @property
    def as_tuple(self) -> Tuple[int, int, str, Optional[str]]:
        """
        Return a tuple of the metric's timestamp, value, measurement and queue.
        """
        return (
            round(self.timestamp),
            self.value,
            self.measurement,
            self.queue_name,
        )

    @classmethod
    def new(cls, start_ms: int, **kwargs) -> "Metric":
        """
        Create a new Metric instance.

        start_ms:
            The start time the request or background job in milliseconds.
        """
        now = datetime.now(timezone.utc).timestamp()
        latency_ms = max(0, int(now * 1000) - start_ms)
        return cls(timestamp=now, value=latency_ms, **kwargs)

    @classmethod
    def for_web(cls, header_value: str) -> Optional["Metric"]:
        """
        Parse the X-Request-Start header value and return a Metric instance.

        Removes non-digits:
            This removes the "t=" prefix added by some web servers (NGINX).
        Removes decimal:
            NGINX also reports this time as fractional seconds with millisecond
            resolution, so removing the decimal gives us integer milliseconds
            (same as Heroku).

        Calculate how long a request has been waiting to be handled, log and
        return a Metric instance.
        """
        request_start = re.sub(r"\D", "", header_value)

        if len(request_start) == 0:
            return None

        logger.debug(f"START X {request_start}")
        metric = Metric.new(start_ms=int(request_start))
        logger.debug(f"queue_time={metric.value}ms")
        return metric

    @classmethod
    def for_queue(cls, queue_name: str, oldest_job_ts: float) -> "Metric":
        """
        Calculate how long the oldest job in a queue has been waiting to be
        handled, log and return a Metric instance.

        queue_name:
            The name of the queue.

        oldest_job_ts:
            The Unix timestamp of the oldest job in the queue.
        """
        metric = Metric.new(start_ms=int(oldest_job_ts * 1000), queue_name=queue_name)
        logger.debug(f"queue_name={queue_name}, queue_time={metric.value}ms")
        return metric



import requests

from judoscale.core.config import config
from judoscale.core.logger import logger


class AdapterApiClient:
    def post_report(self, report):
        try:
            url_metrics = f"{config.api_base_url}/v3/reports"
            metrics_length = len(report["metrics"])
            pid = report["pid"]
            logger.debug(
                f"Posting {metrics_length} metrics from {pid} "
                f"to Judoscale adapter API {url_metrics}"
            )
            requests.post(url_metrics, timeout=5, json=report)
        except requests.RequestException as e:
            logger.warning("Adapter API request failed - {}".format(e))


api_client = AdapterApiClient()



from typing import Optional

from flask import Flask
from flask import __version__ as flask_version
from flask import request

from judoscale.core.adapter import Adapter, AdapterInfo
from judoscale.core.config import config as judoconfig
from judoscale.core.metric import Metric
from judoscale.core.metrics_collectors import WebMetricsCollector
from judoscale.core.reporter import reporter


def store_request_metrics(collector: WebMetricsCollector):
    def inner():
        request_start_header = request.headers.get("x-request-start", "")
        if metric := Metric.for_web(request_start_header):
            collector.add(metric)
        reporter.ensure_running()
        return None

    return inner


class Judoscale:
    def __init__(self, app: Optional[Flask] = None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        judoconfig.merge(app.config.get("JUDOSCALE", {}))
        collector = WebMetricsCollector(judoconfig)
        adapter = Adapter(
            identifier="judoscale-flask",
            adapter_info=AdapterInfo(platform_version=flask_version),
            metrics_collector=collector,
        )
        reporter.add_adapter(adapter)
        reporter.ensure_running()
        app.before_request(store_request_metrics(collector))



import logging
import os

from django.apps import AppConfig
from django.conf import settings
from redis import Redis

from judoscale.core.config import config as judoconfig
from judoscale.core.reporter import reporter
from judoscale.rq import judoscale_rq

logger = logging.getLogger(__name__)


class JudoscaleRQConfig(AppConfig):
    name = "judoscale.rq"
    label = "judoscale_rq"
    verbose_name = "Judoscale (RQ)"

    def ready(self):
        judoconfig.merge(getattr(settings, "JUDOSCALE", {}))

        if judoconfig.api_base_url is None:
            logger.info("Not activated - No API URL provided")
            return

        judoscale_rq(self.redis_connection)
        reporter.ensure_running()

    @property
    def redis_connection(self):
        """
        Return a Redis connection from an RQ queue configuration
        """
        if redis_config := settings.JUDOSCALE.get("REDIS"):
            config = {k.lower(): v for k, v in redis_config.items()}
            if redis_url := config.get("url"):
                return Redis.from_url(redis_url)
            else:
                return Redis(**config)
        elif redis_url := os.getenv("REDIS_URL"):
            return Redis.from_url(redis_url)
        else:
            raise RuntimeError(
                "Missing Redis connection configuration. Please set either "
                "settings.JUDOSCALE['REDIS'] or REDIS_URL environment variable."
            )



import os
import signal
import threading
import time
from platform import python_version
from typing import List

from judoscale.core.adapter import Adapter, AdapterInfo
from judoscale.core.adapter_api_client import api_client
from judoscale.core.config import config
from judoscale.core.logger import logger
from judoscale.core.metric import Metric
from judoscale.core.metrics_collectors import Collector


class Reporter:
    """
    A reporter that collects metrics from all registered collectors and reports
    them to the Judoscale API.
    """

    def __init__(self):
        self._thread = None
        self._running = False
        self._stopevent = threading.Event()
        self.collectors: List[Collector] = []
        self.adapters: List[Adapter] = []

        self.adapters.append(
            Adapter(
                identifier="judoscale-python",
                adapter_info=AdapterInfo(platform_version=python_version()),
            )
        )

    def add_adapter(self, adapter: Adapter):
        """
        Add an adapter to the reporter.

        If the adapter has a metrics collector, it will be added to the list of
        collectors.
        """
        self.adapters.append(adapter)

        if adapter.metrics_collector:
            self.collectors.append(adapter.metrics_collector)

    def start(self):
        logger.info(f"Starting reporter for process {self.pid}")
        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._thread.start()
        self._running = True

    def ensure_running(self):
        try:
            if not self.is_running:
                return self.start()
        except Exception as e:
            logger.warning(f"{e.args} - No reporter has initiated")
            pass

    def signal_handler(self, signum, frame):
        self._stopevent.set()
        self._running = False

    @property
    def is_running(self):
        if self._thread and self._thread.is_alive():
            self._running = True
        else:
            self._running = False
        return self._running

    @property
    def pid(self) -> int:
        return os.getpid()

    def _run_loop(self):
        while self.is_running:
            self._report_metrics()
            time.sleep(config.report_interval_seconds)

            if self._stopevent.is_set():
                break

        logger.info(f"{self.pid} reports completed before exiting.")

    @property
    def all_metrics(self) -> List[Metric]:
        """
        Return a list of all metrics collected by all collectors.
        """
        metrics = []
        for collector in self.collectors:
            metrics.extend(collector.collect())
        return metrics

    def _report_metrics(self) -> None:
        api_client.post_report(self._build_report(self.all_metrics))

    def _build_report(self, metrics: List[Metric]):
        return {
            "dyno": config.dyno,
            "pid": self.pid,
            "config": config.for_report(),
            "adapters": dict(adapter.as_tuple for adapter in self.adapters),
            "metrics": [metric.as_tuple for metric in metrics],
        }


reporter = Reporter()
signal.signal(signal.SIGTERM, reporter.signal_handler)
